@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -66,17 +66,18 @@ struct sim_fs_op {
 	int length;
 	int record_length;
 	int current;
+	unsigned char path[6];
+	unsigned char path_len;
 	gconstpointer cb;
 	gboolean is_read;
 	void *userdata;
 	struct ofono_sim_context *context;
 };
 
-static void sim_fs_op_free(struct sim_fs_op *node)
-{
-	g_free(node->buffer);
-	g_free(node);
-}
+struct ofono_sim_context {
+	struct sim_fs *fs;
+	struct ofono_watchlist *file_watches;
+};
 
 struct sim_fs {
 	GQueue *op_q;
@@ -86,8 +87,19 @@ struct sim_fs {
 	struct ofono_sim *sim;
 	const struct ofono_sim_driver *driver;
 	GSList *contexts;
+	struct ofono_sim_aid_session *session;
+	int session_id;
+	unsigned int watch_id;
 };
 
+static void sim_fs_op_free(gpointer pointer)
+{
+	struct sim_fs_op *node = pointer;
+
+	g_free(node->buffer);
+	g_free(node);
+}
+
 void sim_fs_free(struct sim_fs *fs)
 {
 	if (fs == NULL)
@@ -103,19 +115,15 @@ void sim_fs_free(struct sim_fs *fs)
 	 * for operations still in progress
 	 */
 	if (fs->op_q) {
-		g_queue_foreach(fs->op_q, (GFunc) sim_fs_op_free, NULL);
-		g_queue_free(fs->op_q);
+		g_queue_free_full(fs->op_q, sim_fs_op_free);
+		fs->op_q = NULL;
 	}
 
-	if (fs->contexts != NULL) {
-		GSList *l;
+	while (fs->contexts)
+		sim_fs_context_free(fs->contexts->data);
 
-		for (l = fs->contexts; l; l = l->next) {
-			struct ofono_sim_context *context = l->data;
-
-			sim_fs_context_free(context);
-		}
-	}
+	if (fs->watch_id)
+		__ofono_sim_remove_session_watch(fs->session, fs->watch_id);
 
 	g_free(fs);
 }
@@ -125,11 +133,6 @@ struct file_watch {
 	int ef;
 };
 
-struct ofono_sim_context {
-	struct sim_fs *fs;
-	struct ofono_watchlist *file_watches;
-};
-
 struct sim_fs *sim_fs_new(struct ofono_sim *sim,
 				const struct ofono_sim_driver *driver)
 {
@@ -160,27 +163,47 @@ struct ofono_sim_context *sim_fs_context_new(struct sim_fs *fs)
 	return context;
 }
 
+struct ofono_sim_context *sim_fs_context_new_with_aid(struct sim_fs *fs,
+		unsigned char *aid)
+{
+	struct ofono_sim_context *context = sim_fs_context_new(fs);
+
+	if (context == NULL)
+		return NULL;
+
+	context->fs->session = __ofono_sim_get_session_by_aid(fs->sim, aid);
+	if (!context->fs->session) {
+		sim_fs_context_free(context);
+		return NULL;
+	}
+
+	return context;
+}
+
 void sim_fs_context_free(struct ofono_sim_context *context)
 {
 	struct sim_fs *fs = context->fs;
 	int n = 0;
 	struct sim_fs_op *op;
 
-	while ((op = g_queue_peek_nth(fs->op_q, n)) != NULL) {
-		if (op->context != context) {
-			n += 1;
-			continue;
+	if (fs->op_q) {
+		while ((op = g_queue_peek_nth(fs->op_q, n)) != NULL) {
+			if (op->context != context) {
+				n += 1;
+				continue;
+			}
+
+			if (n == 0) {
+				op->cb = NULL;
+				op->context = NULL;
+
+				n += 1;
+				continue;
+			}
+
+			sim_fs_op_free(op);
+			g_queue_remove(fs->op_q, op);
 		}
-
-		if (n == 0) {
-			op->cb = NULL;
-
-			n += 1;
-			continue;
-		}
-
-		sim_fs_op_free(op);
-		g_queue_remove(fs->op_q, op);
 	}
 
 	if (context->file_watches)
@@ -228,6 +251,9 @@ void sim_fs_notify_file_watches(struct sim_fs *fs, int id)
 		struct ofono_sim_context *context = l->data;
 		GSList *k;
 
+		if (context->file_watches == NULL)
+			continue;
+
 		for (k = context->file_watches->items; k; k = k->next) {
 			struct file_watch *w = k->data;
 			ofono_sim_file_changed_cb_t notify = w->item.notify;
@@ -245,6 +271,8 @@ static void sim_fs_end_current(struct sim_fs *fs)
 
 	if (g_queue_get_length(fs->op_q) > 0)
 		fs->op_source = g_idle_add(sim_fs_op_next, fs);
+	else if (fs->watch_id) /* release the session if no pending reads */
+		__ofono_sim_remove_session_watch(fs->session, fs->watch_id);
 
 	if (fs->fd != -1) {
 		TFR(close(fs->fd));
@@ -476,6 +504,8 @@ static gboolean sim_fs_op_read_block(gpointer user_data)
 	fs->driver->read_file_transparent(fs->sim, op->id,
 						op->current * 256,
 						read_bytes,
+						op->path_len ? op->path : NULL,
+						op->path_len,
 						sim_fs_op_read_block_cb, fs);
 
 	return FALSE;
@@ -565,6 +595,7 @@ static gboolean sim_fs_op_read_record(gpointer user)
 
 		driver->read_file_linear(fs->sim, op->id, op->current,
 						op->record_length,
+						NULL, 0,
 						sim_fs_op_retrieve_cb, fs);
 		break;
 	case OFONO_SIM_FILE_STRUCTURE_CYCLIC:
@@ -575,6 +606,7 @@ static gboolean sim_fs_op_read_record(gpointer user)
 
 		driver->read_file_cyclic(fs->sim, op->id, op->current,
 						op->record_length,
+						NULL, 0,
 						sim_fs_op_retrieve_cb, fs);
 		break;
 	default:
@@ -800,6 +832,99 @@ error:
 	return FALSE;
 }
 
+static void sim_fs_read_session_cb(const struct ofono_error *error,
+		const unsigned char *sdata, int length, void *data)
+{
+	struct sim_fs *fs = data;
+	struct sim_fs_op *op = g_queue_peek_head(fs->op_q);
+	ofono_sim_file_read_cb_t cb;
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
+		sim_fs_op_error(fs);
+		return;
+	}
+
+	cb = op->cb;
+	cb(TRUE, length, 0, sdata, length, op->userdata);
+
+	sim_fs_end_current(fs);
+}
+
+static void session_read_info_cb(const struct ofono_error *error,
+					int filelength,
+					enum ofono_sim_file_structure structure,
+					int recordlength,
+					const unsigned char access[3],
+					unsigned char file_status,
+					void *data)
+{
+	struct sim_fs *fs = data;
+	struct sim_fs_op *op = g_queue_peek_head(fs->op_q);
+
+	if (error->type != OFONO_ERROR_TYPE_NO_ERROR) {
+		sim_fs_op_error(fs);
+		return;
+	}
+
+	sim_fs_op_cache_fileinfo(fs, error, filelength, structure, recordlength,
+			access, file_status);
+
+	if (op->info_only) {
+		sim_fs_read_info_cb_t cb = op->cb;
+
+		cb(1, file_status, filelength, recordlength, op->userdata);
+
+		sim_fs_end_current(fs);
+		return;
+	}
+
+	if (op->structure == OFONO_SIM_FILE_STRUCTURE_TRANSPARENT) {
+		if (!fs->driver->session_read_binary) {
+			sim_fs_op_error(fs);
+			return;
+		}
+
+		fs->driver->session_read_binary(fs->sim, fs->session_id,
+				op->id, op->offset, filelength, op->path,
+				op->path_len, sim_fs_read_session_cb, fs);
+	} else {
+		if (!fs->driver->session_read_record) {
+			sim_fs_op_error(fs);
+			return;
+		}
+
+		fs->driver->session_read_record(fs->sim, fs->session_id,
+				op->id, op->offset, recordlength, op->path,
+				op->path_len, sim_fs_read_session_cb, fs);
+	}
+}
+
+static void session_destroy_cb(void *userdata)
+{
+	struct sim_fs *fs = userdata;
+
+	fs->watch_id = 0;
+}
+
+static void get_session_cb(ofono_bool_t active, int session_id,
+		void *data)
+{
+	struct sim_fs *fs = data;
+	struct sim_fs_op *op;
+
+	if (!active) {
+		sim_fs_op_error(fs);
+		return;
+	}
+
+	op = g_queue_peek_head(fs->op_q);
+
+	fs->session_id = session_id;
+
+	fs->driver->session_read_info(fs->sim, session_id, op->id, op->path,
+			op->path_len, session_read_info_cb, fs);
+}
+
 static gboolean sim_fs_op_next(gpointer user_data)
 {
 	struct sim_fs *fs = user_data;
@@ -822,23 +947,38 @@ static gboolean sim_fs_op_next(gpointer user_data)
 		if (sim_fs_op_check_cached(fs))
 			return FALSE;
 
-		driver->read_file_info(fs->sim, op->id, sim_fs_op_info_cb, fs);
+		if (!fs->session) {
+			driver->read_file_info(fs->sim, op->id,
+						op->path_len ? op->path : NULL,
+						op->path_len,
+						sim_fs_op_info_cb, fs);
+		} else {
+			if (fs->watch_id)
+				fs->driver->session_read_info(fs->sim,
+						fs->session_id, op->id,
+						op->path, op->path_len,
+						session_read_info_cb, fs);
+			else
+				fs->watch_id = __ofono_sim_add_session_watch(
+						fs->session, get_session_cb,
+						fs, session_destroy_cb);
+		}
 	} else {
 		switch (op->structure) {
 		case OFONO_SIM_FILE_STRUCTURE_TRANSPARENT:
 			driver->write_file_transparent(fs->sim, op->id, 0,
 					op->length, op->buffer,
-					sim_fs_op_write_cb, fs);
+					NULL, 0, sim_fs_op_write_cb, fs);
 			break;
 		case OFONO_SIM_FILE_STRUCTURE_FIXED:
 			driver->write_file_linear(fs->sim, op->id, op->current,
 					op->length, op->buffer,
-					sim_fs_op_write_cb, fs);
+					NULL, 0, sim_fs_op_write_cb, fs);
 			break;
 		case OFONO_SIM_FILE_STRUCTURE_CYCLIC:
 			driver->write_file_cyclic(fs->sim, op->id,
 					op->length, op->buffer,
-					sim_fs_op_write_cb, fs);
+					NULL, 0, sim_fs_op_write_cb, fs);
 			break;
 		default:
 			ofono_error("Unrecognized file structure, "
@@ -894,6 +1034,7 @@ int sim_fs_read_info(struct ofono_sim_context *context, int id,
 int sim_fs_read(struct ofono_sim_context *context, int id,
 		enum ofono_sim_file_structure expected_type,
 		unsigned short offset, unsigned short num_bytes,
+		const unsigned char *path, unsigned int path_len,
 		ofono_sim_file_read_cb_t cb, void *data)
 {
 	struct sim_fs *fs = context->fs;
@@ -905,8 +1046,18 @@ int sim_fs_read(struct ofono_sim_context *context, int id,
 	if (fs->driver == NULL)
 		return -EINVAL;
 
-	if (fs->driver->read_file_info == NULL)
-		return -ENOSYS;
+	/* check driver support for session based read */
+	if (fs->session) {
+		if (!fs->driver->session_read_info) {
+			cb(0, 0, 0, NULL, 0, data);
+			return -ENOSYS;
+		}
+	} else {
+		if (!fs->driver->read_file_info) {
+			cb(0, 0, 0, NULL, 0, data);
+			return -ENOSYS;
+		}
+	}
 
 	if (fs->op_q == NULL)
 		fs->op_q = g_queue_new();
@@ -924,6 +1075,8 @@ int sim_fs_read(struct ofono_sim_context *context, int id,
 	op->num_bytes = num_bytes;
 	op->info_only = FALSE;
 	op->context = context;
+	memcpy(op->path, path, path_len);
+	op->path_len = path_len;
 
 	g_queue_push_tail(fs->op_q, op);
 
